package sun.font;

import java.awt.geom.NoninvertibleTransformException;
import java.awt.Point;
import java.awt.font.GlyphJustificationInfo;
import java.awt.font.GlyphMetrics;
import java.awt.geom.GeneralPath;
import java.awt.Shape;
import java.awt.Rectangle;
import java.awt.font.LineMetrics;
import java.awt.geom.Rectangle2D;
import java.awt.RenderingHints;
import sun.java2d.loops.FontInfo;
import java.text.CharacterIterator;
import java.awt.geom.Point2D;
import java.lang.ref.SoftReference;
import java.awt.geom.AffineTransform;
import java.awt.font.FontRenderContext;
import java.awt.Font;
import java.awt.font.GlyphVector;

public class StandardGlyphVector extends GlyphVector
{
    private Font font;
    private FontRenderContext frc;
    private int[] glyphs;
    private int[] userGlyphs;
    private float[] positions;
    private int[] charIndices;
    private int flags;
    private static final int UNINITIALIZED_FLAGS = -1;
    private StandardGlyphVector.GlyphTransformInfo gti;
    private AffineTransform ftx;
    private AffineTransform dtx;
    private AffineTransform invdtx;
    private AffineTransform frctx;
    private Font2D font2D;
    private SoftReference fsref;
    private SoftReference lbcacheRef;
    private SoftReference vbcacheRef;
    public static final int FLAG_USES_VERTICAL_BASELINE = 128;
    public static final int FLAG_USES_VERTICAL_METRICS = 256;
    public static final int FLAG_USES_ALTERNATE_ORIENTATION = 512;
    
    public StandardGlyphVector(Font font, String s, FontRenderContext fontRenderContext) {
        this.init(font, s.toCharArray(), 0, s.length(), fontRenderContext, -1);
    }
    
    public StandardGlyphVector(Font font, char[] array, FontRenderContext fontRenderContext) {
        this.init(font, array, 0, array.length, fontRenderContext, -1);
    }
    
    public StandardGlyphVector(Font font, char[] array, int n, int n2, FontRenderContext fontRenderContext) {
        this.init(font, array, n, n2, fontRenderContext, -1);
    }
    
    private float getTracking(Font font) {
        if (font.hasLayoutAttributes()) {
            return ((AttributeMap)font.getAttributes()).getValues().getTracking();
        }
        return 0.0f;
    }
    
    public StandardGlyphVector(Font font, FontRenderContext fontRenderContext, int[] array, float[] array2, int[] array3, int n) {
        this.initGlyphVector(font, fontRenderContext, array, array2, array3, n);
        float tracking = this.getTracking(font);
        if (tracking != 0.0f) {
            Point2D.Float float1 = new Point2D.Float(tracking * font.getSize2D(), 0.0f);
            if (font.isTransformed()) {
                font.getTransform().deltaTransform(float1, float1);
            }
            FontStrike strike = FontUtilities.getFont2D(font).getStrike(font, fontRenderContext);
            float[] array4 = { float1.x, float1.y };
            for (int i = 0; i < array4.length; ++i) {
                float n2 = array4[i];
                if (n2 != 0.0f) {
                    float n3 = 0.0f;
                    for (int j = 0; j < array.length; i += 2) {
                        if (strike.getGlyphAdvance(array[j++]) != 0.0f) {
                            int n4 = i;
                            array2[n4] += n3;
                            n3 += n2;
                        }
                    }
                    int n5 = array2.length - 2 + i;
                    array2[n5] += n3;
                }
            }
        }
    }
    
    public void initGlyphVector(Font font, FontRenderContext frc, int[] array, float[] positions, int[] charIndices, int flags) {
        this.font = font;
        this.frc = frc;
        this.glyphs = array;
        this.userGlyphs = array;
        this.positions = positions;
        this.charIndices = charIndices;
        this.flags = flags;
        this.initFontData();
    }
    
    public StandardGlyphVector(Font font, CharacterIterator characterIterator, FontRenderContext fontRenderContext) {
        int beginIndex = characterIterator.getBeginIndex();
        char[] array = new char[characterIterator.getEndIndex() - beginIndex];
        for (int i = characterIterator.first(); i != '\uffff'; i = characterIterator.next()) {
            array[characterIterator.getIndex() - beginIndex] = i;
        }
        this.init(font, array, 0, array.length, fontRenderContext, -1);
    }
    
    public StandardGlyphVector(Font font, int[] userGlyphs, FontRenderContext frc) {
        this.font = font;
        this.frc = frc;
        this.flags = -1;
        this.initFontData();
        this.userGlyphs = userGlyphs;
        this.glyphs = this.getValidatedGlyphs(this.userGlyphs);
    }
    
    public static StandardGlyphVector getStandardGV(GlyphVector glyphVector, FontInfo fontInfo) {
        if (fontInfo.aaHint == 2) {
            Object antiAliasingHint = glyphVector.getFontRenderContext().getAntiAliasingHint();
            if (antiAliasingHint != RenderingHints.VALUE_TEXT_ANTIALIAS_ON && antiAliasingHint != RenderingHints.VALUE_TEXT_ANTIALIAS_GASP) {
                FontRenderContext fontRenderContext = glyphVector.getFontRenderContext();
                return new StandardGlyphVector(glyphVector, new FontRenderContext(fontRenderContext.getTransform(), RenderingHints.VALUE_TEXT_ANTIALIAS_ON, fontRenderContext.getFractionalMetricsHint()));
            }
        }
        if ((glyphVector instanceof StandardGlyphVector)) {
            return (StandardGlyphVector)glyphVector;
        }
        return new StandardGlyphVector(glyphVector, glyphVector.getFontRenderContext());
    }
    
    @Override
    public Font getFont() {
        return this.font;
    }
    
    @Override
    public FontRenderContext getFontRenderContext() {
        return this.frc;
    }
    
    @Override
    public void performDefaultLayout() {
        this.positions = null;
        if (this.getTracking(this.font) == 0.0f) {
            this.clearFlags(2);
        }
    }
    
    @Override
    public int getNumGlyphs() {
        return this.glyphs.length;
    }
    
    @Override
    public int getGlyphCode(int n) {
        return this.userGlyphs[n];
    }
    
    @Override
    public int[] getGlyphCodes(int i, int j, int[] array) {
        if (j < 0) {
            throw new IllegalArgumentException("count = " + j);
        }
        if (i < 0) {
            throw new IndexOutOfBoundsException("start = " + i);
        }
        if (i > this.glyphs.length - j) {
            throw new IndexOutOfBoundsException("start + count = " + (i + j));
        }
        if (array == null) {
            array = new int[j];
        }
        for (int k = 0; k < j; ++k) {
            array[k] = this.userGlyphs[k + i];
        }
        return array;
    }
    
    @Override
    public int getGlyphCharIndex(int i) {
        if (i < 0 && i >= this.glyphs.length) {
            throw new IndexOutOfBoundsException("" + i);
        }
        if (this.charIndices == null) {
            if ((this.getLayoutFlags() & 0x4) != 0x0) {
                return this.glyphs.length - 1 - i;
            }
            return i;
        }
        else {
            return this.charIndices[i];
        }
    }
    
    @Override
    public int[] getGlyphCharIndices(int i, int j, int[] array) {
        if (i < 0 || j < 0 || j > this.glyphs.length - i) {
            throw new IndexOutOfBoundsException("" + i + ", " + j);
        }
        if (array == null) {
            array = new int[j];
        }
        if (this.charIndices == null) {
            if ((this.getLayoutFlags() & 0x4) != 0x0) {
                for (int n = 0, n2 = this.glyphs.length - 1 - i; n < j; ++n, --n2) {
                    array[n] = n2;
                }
            }
            else {
                int l;
                for (l = 0, ; l < j; ++l, ++i) {
                    array[l] = i;
                }
            }
        }
        else {
            for (int k = 0; k < j; ++k) {
                array[k] = this.charIndices[k + i];
            }
        }
        return array;
    }
    
    @Override
    public Rectangle2D getLogicalBounds() {
        this.setFRCTX();
        this.initPositions();
        LineMetrics lineMetrics = this.font.getLineMetrics("", this.frc);
        float x = 0.0f;
        float y = -lineMetrics.getAscent();
        float n = 0.0f;
        float n2 = lineMetrics.getDescent() + lineMetrics.getLeading();
        if (this.glyphs.length > 0) {
            n = this.positions[this.positions.length - 2];
        }
        return new Rectangle2D.Float(x, y, n - x, n2 - y);
    }
    
    @Override
    public Rectangle2D getVisualBounds() {
        Object o = null;
        for (int i = 0; i < this.glyphs.length; ++i) {
            Rectangle2D bounds2D = this.getGlyphVisualBounds(i).getBounds2D();
            if (!bounds2D.isEmpty()) {
                if (o == null) {
                    o = bounds2D;
                }
                else {
                    Rectangle2D.union(o, bounds2D, o);
                }
            }
        }
        if (o == null) {
            o = new Rectangle2D.Float(0.0f, 0.0f, 0.0f, 0.0f);
            return o;
        }
        return o;
    }
    
    @Override
    public Rectangle getPixelBounds(FontRenderContext fontRenderContext, float n, float n2) {
        return this.getGlyphsPixelBounds(fontRenderContext, n, n2, 0, this.glyphs.length);
    }
    
    @Override
    public Shape getOutline() {
        return this.getGlyphsOutline(0, this.glyphs.length, 0.0f, 0.0f);
    }
    
    @Override
    public Shape getOutline(float n, float n2) {
        return this.getGlyphsOutline(0, this.glyphs.length, n, n2);
    }
    
    @Override
    public Shape getGlyphOutline(int n) {
        return this.getGlyphsOutline(n, 1, 0.0f, 0.0f);
    }
    
    @Override
    public Shape getGlyphOutline(int n, float n2, float n3) {
        return this.getGlyphsOutline(n, 1, n2, n3);
    }
    
    @Override
    public Point2D getGlyphPosition(int n) {
        this.initPositions();
        n *= 2;
        return new Point2D.Float(this.positions[n], this.positions[n + 1]);
    }
    
    @Override
    public void setGlyphPosition(int n, Point2D point2D) {
        this.initPositions();
        int n2 = n << 1;
        this.positions[n2] = (float)point2D.getX();
        this.positions[n2 + 1] = (float)point2D.getY();
        this.clearCaches(n);
        this.addFlags(2);
    }
    
    @Override
    public AffineTransform getGlyphTransform(int i) {
        if (i < 0 || i >= this.glyphs.length) {
            throw new IndexOutOfBoundsException("ix = " + i);
        }
        if (this.gti != null) {
            return this.gti.getGlyphTransform(i);
        }
        return null;
    }
    
    @Override
    public void setGlyphTransform(int i, AffineTransform affineTransform) {
        if (i < 0 || i >= this.glyphs.length) {
            throw new IndexOutOfBoundsException("ix = " + i);
        }
        if (this.gti == null) {
            if (affineTransform == null || affineTransform.isIdentity()) {
                return;
            }
            this.gti = new StandardGlyphVector.GlyphTransformInfo(this);
        }
        this.gti.setGlyphTransform(i, affineTransform);
        if (this.gti.transformCount() == 0) {
            this.gti = null;
        }
        return;
    }
    
    @Override
    public int getLayoutFlags() {
        if (this.flags == -1) {
            this.flags = 0;
            if (this.charIndices != null && this.glyphs.length > 1) {
                boolean n = true;
                boolean n2 = true;
                int n4;
                for (int length = this.charIndices.length, n3 = 0; n3 < this.charIndices.length && (n || n2); n = (n && n4 == n3), n2 = (n2 && n4 == --length), ++n3) {
                    n4 = this.charIndices[n3];
                }
                if (n2 != 0) {
                    this.flags |= 0x4;
                }
                if (n2 == 0 && n == 0) {
                    this.flags |= 0x8;
                }
            }
        }
        return this.flags;
    }
    
    @Override
    public float[] getGlyphPositions(int i, int j, float[] array) {
        if (j < 0) {
            throw new IllegalArgumentException("count = " + j);
        }
        if (i < 0) {
            throw new IndexOutOfBoundsException("start = " + i);
        }
        if (i > this.glyphs.length + 1 - j) {
            throw new IndexOutOfBoundsException("start + count = " + (i + j));
        }
        return this.internalGetGlyphPositions(i, j, 0, array);
    }
    
    @Override
    public Shape getGlyphLogicalBounds(int i) {
        if (i < 0 || i >= this.glyphs.length) {
            throw new IndexOutOfBoundsException("ix = " + i);
        }
        Shape[] referent;
        if (this.lbcacheRef == null || (referent = (Shape[])(Shape[])this.lbcacheRef.get()) == null) {
            referent = new Shape[this.glyphs.length];
            this.lbcacheRef = new SoftReference(referent);
        }
        Shape shape = referent[i];
        if (shape == null) {
            this.setFRCTX();
            this.initPositions();
            StandardGlyphVector.ADL o = new StandardGlyphVector.ADL();
            StandardGlyphVector.GlyphStrike glyphStrike = this.getGlyphStrike(i);
            glyphStrike.getADL(o);
            Point2D.Float glyphMetrics = glyphStrike.strike.getGlyphMetrics(this.glyphs[i]);
            float x = glyphMetrics.x;
            float y = glyphMetrics.y;
            float n = o.descentX + o.leadingX + o.ascentX;
            float n2 = o.descentY + o.leadingY + o.ascentY;
            float x2 = this.positions[i * 2] + glyphStrike.dx - o.ascentX;
            float y2 = this.positions[i * 2 + 1] + glyphStrike.dy - o.ascentY;
            GeneralPath o2 = new GeneralPath();
            o2.moveTo(x2, y2);
            o2.lineTo(x2 + x, y2 + y);
            o2.lineTo(x2 + x + n, y2 + y + n2);
            o2.lineTo(x2 + n, y2 + n2);
            o2.closePath();
            shape = new DelegatingShape(o2);
            referent[i] = shape;
            return shape;
        }
        return shape;
    }
    
    @Override
    public Shape getGlyphVisualBounds(int i) {
        if (i < 0 || i >= this.glyphs.length) {
            throw new IndexOutOfBoundsException("ix = " + i);
        }
        Shape[] referent;
        if (this.vbcacheRef == null || (referent = (Shape[])(Shape[])this.vbcacheRef.get()) == null) {
            referent = new Shape[this.glyphs.length];
            this.vbcacheRef = new SoftReference(referent);
        }
        Shape shape = referent[i];
        if (shape == null) {
            shape = new DelegatingShape(this.getGlyphOutlineBounds(i));
            referent[i] = shape;
            return shape;
        }
        return shape;
    }
    
    @Override
    public Rectangle getGlyphPixelBounds(int n, FontRenderContext fontRenderContext, float n2, float n3) {
        return this.getGlyphsPixelBounds(fontRenderContext, n2, n3, n, 1);
    }
    
    @Override
    public GlyphMetrics getGlyphMetrics(int i) {
        if (i < 0 || i >= this.glyphs.length) {
            throw new IndexOutOfBoundsException("ix = " + i);
        }
        Rectangle2D bounds2D = this.getGlyphVisualBounds(i).getBounds2D();
        Point2D glyphPosition = this.getGlyphPosition(i);
        bounds2D.setRect(bounds2D.getMinX() - glyphPosition.getX(), bounds2D.getMinY() - glyphPosition.getY(), bounds2D.getWidth(), bounds2D.getHeight());
        Point2D.Float glyphMetrics = this.getGlyphStrike(i).strike.getGlyphMetrics(this.glyphs[i]);
        return new GlyphMetrics(true, glyphMetrics.x, glyphMetrics.y, bounds2D, (byte)0);
    }
    
    @Override
    public GlyphJustificationInfo getGlyphJustificationInfo(int i) {
        if (i < 0 || i >= this.glyphs.length) {
            throw new IndexOutOfBoundsException("ix = " + i);
        }
        return null;
    }
    
    @Override
    public boolean equals(GlyphVector glyphVector) {
        if (this == glyphVector) {
            return true;
        }
        if (glyphVector == null) {
            return false;
        }
        try {
            StandardGlyphVector standardGlyphVector = (StandardGlyphVector)glyphVector;
            if (this.glyphs.length != standardGlyphVector.glyphs.length) {
                return false;
            }
            int i = 0;
            while (i < this.glyphs.length) {
                if (this.glyphs[i] != standardGlyphVector.glyphs[i]) {
                    return false;
                }
                ++i;
            }
            if (!this.font.equals(standardGlyphVector.font)) {
                return false;
            }
            if (!this.frc.equals(standardGlyphVector.frc)) {
                return false;
            }
            if (standardGlyphVector.positions == null != (this.positions == null)) {
                if (this.positions == null) {
                    this.initPositions();
                }
                else {
                    standardGlyphVector.initPositions();
                }
            }
            if (this.positions != null) {
                int j = 0;
                while (j < this.positions.length) {
                    if (this.positions[j] != standardGlyphVector.positions[j]) {
                        return false;
                    }
                    ++j;
                }
            }
            if (this.gti == null) {
                return standardGlyphVector.gti == null;
            }
            return this.gti.equals(standardGlyphVector.gti);
        }
        catch (ClassCastException ex) {
            return false;
        }
    }
    
    @Override
    public int hashCode() {
        return this.font.hashCode() ^ this.glyphs.length;
    }
    
    @Override
    public boolean equals(Object o) {
        try {
            return this.equals((GlyphVector)o);
        }
        catch (ClassCastException ex) {
            return false;
        }
    }
    
    public StandardGlyphVector copy() {
        return (StandardGlyphVector)this.clone();
    }
    
    public Object clone() {
        try {
            StandardGlyphVector standardGlyphVector = (StandardGlyphVector)super.clone();
            standardGlyphVector.clearCaches();
            if (this.positions != null) {
                standardGlyphVector.positions = (float[])(float[])this.positions.clone();
            }
            if (this.gti != null) {
                standardGlyphVector.gti = new StandardGlyphVector.GlyphTransformInfo(standardGlyphVector, this.gti);
                return standardGlyphVector;
            }
            return standardGlyphVector;
        }
        catch (CloneNotSupportedException ex) {
            return this;
        }
    }
    
    public void setGlyphPositions(float[] array, int n, int n2, int i) {
        if (i < 0) {
            throw new IllegalArgumentException("count = " + i);
        }
        this.initPositions();
        int j;
        int n3;
        for (j = n2 * 2, n3 = j + i * 2, ; j < n3; ++j, ++n) {
            this.positions[j] = array[n];
        }
        this.clearCaches();
        this.addFlags(2);
        return;
    }
    
    public void setGlyphPositions(float[] array) {
        int i = this.glyphs.length * 2 + 2;
        if (array.length != i) {
            throw new IllegalArgumentException("srcPositions.length != " + i);
        }
        this.positions = (float[])(float[])array.clone();
        this.clearCaches();
        this.addFlags(2);
        return;
    }
    
    public float[] getGlyphPositions(float[] array) {
        return this.internalGetGlyphPositions(0, this.glyphs.length + 1, 0, array);
    }
    
    public AffineTransform[] getGlyphTransforms(int i, int j, AffineTransform[] array) {
        if (i < 0 || j < 0 || i + j > this.glyphs.length) {
            throw new IllegalArgumentException("start: " + i + " count: " + j);
        }
        if (this.gti == null) {
            return null;
        }
        if (array == null) {
            array = new AffineTransform[j];
        }
        for (int k = 0; k < j; ++k, ++i) {
            array[k] = this.gti.getGlyphTransform(i);
        }
        return array;
    }
    
    public AffineTransform[] getGlyphTransforms() {
        return this.getGlyphTransforms(0, this.glyphs.length, (AffineTransform[])null);
    }
    
    public void setGlyphTransforms(AffineTransform[] array, int n, int i, int n2) {
        for (; i < i + n2; ++i) {
            this.setGlyphTransform(i, array[n + i]);
        }
    }
    
    public void setGlyphTransforms(AffineTransform[] array) {
        this.setGlyphTransforms(array, 0, 0, this.glyphs.length);
    }
    
    public float[] getGlyphInfo() {
        this.setFRCTX();
        this.initPositions();
        float[] array = new float[this.glyphs.length * 8];
        for (int i = 0, n = 0; i < this.glyphs.length; ++i, n += 8) {
            float n2 = this.positions[i * 2];
            float n3 = this.positions[i * 2 + 1];
            array[n] = n2;
            array[n + 1] = n3;
            Point2D.Float glyphMetrics = this.getGlyphStrike(i).strike.getGlyphMetrics(this.glyphs[i]);
            array[n + 2] = glyphMetrics.x;
            array[n + 3] = glyphMetrics.y;
            Rectangle2D bounds2D = this.getGlyphVisualBounds(i).getBounds2D();
            array[n + 4] = (float)bounds2D.getMinX();
            array[n + 5] = (float)bounds2D.getMinY();
            array[n + 6] = (float)bounds2D.getWidth();
            array[n + 7] = (float)bounds2D.getHeight();
        }
        return array;
    }
    
    public void pixellate(FontRenderContext frc, Point2D point2D, Point point) {
        if (frc == null) {
            frc = this.frc;
        }
        AffineTransform transform = frc.getTransform();
        transform.transform(point2D, point2D);
        point.x = (int)point2D.getX();
        point.y = (int)point2D.getY();
        point2D.setLocation((double)point.x, (double)point.y);
        try {
            transform.inverseTransform(point2D, point2D);
        }
        catch (NoninvertibleTransformException ex) {
            throw new IllegalArgumentException("must be able to invert frc transform");
        }
    }
    
    boolean needsPositions(double[] array) {
        return this.gti != null || (this.getLayoutFlags() & 0x2) != 0x0 || !matchTX(array, this.frctx);
    }
    
    Object setupGlyphImages(long[] array, float[] dstPts, double[] renderTransform) {
        this.initPositions();
        this.setRenderTransform(renderTransform);
        if (this.gti != null) {
            return this.gti.setupGlyphImages(array, dstPts, this.dtx);
        }
        StandardGlyphVector.GlyphStrike defaultStrike = this.getDefaultStrike();
        defaultStrike.strike.getGlyphImagePtrs(this.glyphs, array, this.glyphs.length);
        if (dstPts != null) {
            if (this.dtx.isIdentity()) {
                System.arraycopy(this.positions, 0, dstPts, 0, this.glyphs.length * 2);
            }
            else {
                this.dtx.transform(this.positions, 0, dstPts, 0, this.glyphs.length);
            }
        }
        return defaultStrike;
    }
    
    private static boolean matchTX(double[] array, AffineTransform affineTransform) {
        return array[0] == affineTransform.getScaleX() && array[1] == affineTransform.getShearY() && array[2] == affineTransform.getShearX() && array[3] == affineTransform.getScaleY();
    }
    
    private static AffineTransform getNonTranslateTX(AffineTransform affineTransform) {
        if (affineTransform.getTranslateX() != 0.0 || affineTransform.getTranslateY() != 0.0) {
            affineTransform = new AffineTransform(affineTransform.getScaleX(), affineTransform.getShearY(), affineTransform.getShearX(), affineTransform.getScaleY(), 0.0, 0.0);
            return affineTransform;
        }
        return affineTransform;
    }
    
    private static boolean equalNonTranslateTX(AffineTransform affineTransform, AffineTransform affineTransform2) {
        return affineTransform.getScaleX() == affineTransform2.getScaleX() && affineTransform.getShearY() == affineTransform2.getShearY() && affineTransform.getShearX() == affineTransform2.getShearX() && affineTransform.getScaleY() == affineTransform2.getScaleY();
    }
    
    private void setRenderTransform(double[] flatmatrix) {
        assert flatmatrix.length == 4;
        if (!matchTX(flatmatrix, this.dtx)) {
            this.resetDTX(new AffineTransform(flatmatrix));
        }
        return;
    }
    
    private final void setDTX(AffineTransform affineTransform) {
        if (!equalNonTranslateTX(this.dtx, affineTransform)) {
            this.resetDTX(getNonTranslateTX(affineTransform));
        }
    }
    
    private final void setFRCTX() {
        if (!equalNonTranslateTX(this.frctx, this.dtx)) {
            this.resetDTX(getNonTranslateTX(this.frctx));
        }
    }
    
    private final void resetDTX(AffineTransform dtx) {
        this.fsref = null;
        this.dtx = dtx;
        this.invdtx = null;
        if (!this.dtx.isIdentity()) {
            try {
                this.invdtx = this.dtx.createInverse();
            }
            catch (NoninvertibleTransformException ex) {}
        }
        if (this.gti != null) {
            this.gti.strikesRef = null;
        }
    }
    
    private StandardGlyphVector(GlyphVector glyphVector, FontRenderContext frc) {
        this.font = glyphVector.getFont();
        this.frc = frc;
        this.initFontData();
        int numGlyphs = glyphVector.getNumGlyphs();
        this.userGlyphs = glyphVector.getGlyphCodes(0, numGlyphs, (int[])null);
        if ((glyphVector instanceof StandardGlyphVector)) {
            this.glyphs = this.userGlyphs;
        }
        else {
            this.glyphs = this.getValidatedGlyphs(this.userGlyphs);
        }
        this.flags = (glyphVector.getLayoutFlags() & 0xF);
        if ((this.flags & 0x2) != 0x0) {
            this.positions = glyphVector.getGlyphPositions(0, numGlyphs + 1, (float[])null);
        }
        if ((this.flags & 0x8) != 0x0) {
            this.charIndices = glyphVector.getGlyphCharIndices(0, numGlyphs, (int[])null);
        }
        if ((this.flags & 0x1) != 0x0) {
            AffineTransform[] glyphTransforms = new AffineTransform[numGlyphs];
            for (int i = 0; i < numGlyphs; ++i) {
                glyphTransforms[i] = glyphVector.getGlyphTransform(i);
            }
            this.setGlyphTransforms(glyphTransforms);
        }
    }
    
    int[] getValidatedGlyphs(int[] array) {
        int length = array.length;
        int[] array2 = new int[length];
        for (int i = 0; i < length; ++i) {
            if (array[i] == 65534 || array[i] == 65535) {
                array2[i] = array[i];
            }
            else {
                array2[i] = this.font2D.getValidatedGlyphCode(array[i]);
            }
        }
        return array2;
    }
    
    private void init(Font font, char[] array, int n, int n2, FontRenderContext frc, int flags) {
        if (n < 0 || n2 < 0 || n + n2 > array.length) {
            throw new ArrayIndexOutOfBoundsException("start or count out of bounds");
        }
        this.font = font;
        this.frc = frc;
        this.flags = flags;
        if (this.getTracking(font) != 0.0f) {
            this.addFlags(2);
        }
        if (n != 0) {
            char[] array2 = new char[n2];
            System.arraycopy(array, n, array2, 0, n2);
            array = array2;
        }
        this.initFontData();
        this.glyphs = new int[n2];
        this.userGlyphs = this.glyphs;
        this.font2D.getMapper().charsToGlyphs(n2, array, this.glyphs);
        return;
    }
    
    private void initFontData() {
        this.font2D = FontUtilities.getFont2D(this.font);
        if ((this.font2D instanceof FontSubstitution)) {
            this.font2D = ((FontSubstitution)this.font2D).getCompositeFont2D();
        }
        float size2D = this.font.getSize2D();
        if (this.font.isTransformed()) {
            this.ftx = this.font.getTransform();
            if (this.ftx.getTranslateX() != 0.0 || this.ftx.getTranslateY() != 0.0) {
                this.addFlags(2);
            }
            this.ftx.setTransform(this.ftx.getScaleX(), this.ftx.getShearY(), this.ftx.getShearX(), this.ftx.getScaleY(), 0.0, 0.0);
            this.ftx.scale((double)size2D, (double)size2D);
        }
        else {
            this.ftx = AffineTransform.getScaleInstance((double)size2D, (double)size2D);
        }
        this.frctx = this.frc.getTransform();
        this.resetDTX(getNonTranslateTX(this.frctx));
    }
    
    private float[] internalGetGlyphPositions(int n, int n2, int i, float[] array) {
        if (array == null) {
            array = new float[i + n2 * 2];
        }
        this.initPositions();
        for (int n3 = i + n2 * 2, n4 = n * 2; i < n3; ++i, ++n4) {
            array[i] = this.positions[n4];
        }
        return array;
    }
    
    private Rectangle2D getGlyphOutlineBounds(int n) {
        this.setFRCTX();
        this.initPositions();
        return this.getGlyphStrike(n).getGlyphOutlineBounds(this.glyphs[n], this.positions[n * 2], this.positions[n * 2 + 1]);
    }
    
    private Shape getGlyphsOutline(int i, int n, float n2, float n3) {
        this.setFRCTX();
        this.initPositions();
        GeneralPath generalPath = new GeneralPath(1);
        for (int n4 = i + n, n5 = i * 2; i < n4; ++i, n5 += 2) {
            this.getGlyphStrike(i).appendGlyphOutline(this.glyphs[i], generalPath, n2 + this.positions[n5], n3 + this.positions[n5 + 1]);
        }
        return generalPath;
    }
    
    private Rectangle getGlyphsPixelBounds(FontRenderContext fontRenderContext, float n, float n2, int n3, int n4) {
        this.initPositions();
        AffineTransform dtx;
        if (fontRenderContext == null || fontRenderContext.equals(this.frc)) {
            dtx = this.frctx;
        }
        else {
            dtx = fontRenderContext.getTransform();
        }
        this.setDTX(dtx);
        if (this.gti != null) {
            return this.gti.getGlyphsPixelBounds(dtx, n, n2, n3, n4);
        }
        FontStrike strike = this.getDefaultStrike().strike;
        Object o = null;
        Rectangle o2 = new Rectangle();
        Point2D.Float o3 = new Point2D.Float();
        int n5 = n3 * 2;
        while (--n4 >= 0) {
            o3.x = n + this.positions[n5++];
            o3.y = n2 + this.positions[n5++];
            dtx.transform(o3, o3);
            strike.getGlyphImageBounds(this.glyphs[n3++], o3, o2);
            if (o2.isEmpty()) {
                continue;
            }
            if (o == null) {
                o = new Rectangle(o2);
            }
            else {
                o.add(o2);
            }
        }
        return (o != null) ? o : o2;
    }
    
    private void clearCaches(int n) {
        if (this.lbcacheRef != null) {
            Shape[] array = (Shape[])(Shape[])this.lbcacheRef.get();
            if (array != null) {
                array[n] = null;
            }
        }
        if (this.vbcacheRef != null) {
            Shape[] array2 = (Shape[])(Shape[])this.vbcacheRef.get();
            if (array2 != null) {
                array2[n] = null;
            }
        }
    }
    
    private void clearCaches() {
        this.lbcacheRef = null;
        this.vbcacheRef = null;
    }
    
    private void initPositions() {
        if (this.positions == null) {
            this.setFRCTX();
            this.positions = new float[this.glyphs.length * 2 + 2];
            Point2D.Float float1 = null;
            float tracking = this.getTracking(this.font);
            if (tracking != 0.0f) {
                float1 = new Point2D.Float(tracking * this.font.getSize2D(), 0.0f);
            }
            Point2D.Float o = new Point2D.Float(0.0f, 0.0f);
            if (this.font.isTransformed()) {
                AffineTransform transform = this.font.getTransform();
                transform.transform(o, o);
                this.positions[0] = o.x;
                this.positions[1] = o.y;
                if (float1 != null) {
                    transform.deltaTransform(float1, float1);
                }
            }
            for (int i = 0, n = 2; i < this.glyphs.length; this.positions[n] = o.x, this.positions[n + 1] = o.y, ++i, n += 2) {
                this.getGlyphStrike(i).addDefaultGlyphAdvance(this.glyphs[i], o);
                if (float1 != null) {
                    o.x += float1.x;
                    o.y += float1.y;
                }
            }
        }
    }
    
    private void addFlags(int n) {
        this.flags = (this.getLayoutFlags() | n);
    }
    
    private void clearFlags(int n) {
        this.flags = (this.getLayoutFlags() & ~n);
    }
    
    private StandardGlyphVector.GlyphStrike getGlyphStrike(int n) {
        if (this.gti == null) {
            return this.getDefaultStrike();
        }
        return this.gti.getStrike(n);
    }
    
    private StandardGlyphVector.GlyphStrike getDefaultStrike() {
        StandardGlyphVector.GlyphStrike create = null;
        if (this.fsref != null) {
            create = (StandardGlyphVector.GlyphStrike)this.fsref.get();
        }
        if (create == null) {
            create = StandardGlyphVector.GlyphStrike.create(this, this.dtx, null);
            this.fsref = new SoftReference(create);
            return create;
        }
        return create;
    }
    
    @Override
    public String toString() {
        return this.appendString(null).toString();
    }
    
    StringBuffer appendString(StringBuffer sb) {
        if (sb == null) {
            sb = new StringBuffer();
        }
        try {
            sb.append("SGV{font: ");
            sb.append(this.font.toString());
            sb.append(", frc: ");
            sb.append(this.frc.toString());
            sb.append(", glyphs: (");
            sb.append(this.glyphs.length);
            sb.append(")[");
            for (int i = 0; i < this.glyphs.length; ++i) {
                if (i > 0) {
                    sb.append(", ");
                }
                sb.append(Integer.toHexString(this.glyphs[i]));
            }
            sb.append("]");
            if (this.positions != null) {
                sb.append(", positions: (");
                sb.append(this.positions.length);
                sb.append(")[");
                for (int j = 0; j < this.positions.length; j += 2) {
                    if (j > 0) {
                        sb.append(", ");
                    }
                    sb.append(this.positions[j]);
                    sb.append("@");
                    sb.append(this.positions[j + 1]);
                }
                sb.append("]");
            }
            if (this.charIndices != null) {
                sb.append(", indices: (");
                sb.append(this.charIndices.length);
                sb.append(")[");
                for (int k = 0; k < this.charIndices.length; ++k) {
                    if (k > 0) {
                        sb.append(", ");
                    }
                    sb.append(this.charIndices[k]);
                }
                sb.append("]");
            }
            sb.append(", flags:");
            if (this.getLayoutFlags() == 0) {
                sb.append(" default");
            }
            else {
                if ((this.flags & 0x1) != 0x0) {
                    sb.append(" tx");
                }
                if ((this.flags & 0x2) != 0x0) {
                    sb.append(" pos");
                }
                if ((this.flags & 0x4) != 0x0) {
                    sb.append(" rtl");
                }
                if ((this.flags & 0x8) != 0x0) {
                    sb.append(" complex");
                }
            }
        }
        catch (Exception ex) {
            sb.append(" " + ex.getMessage());
        }
        sb.append("}");
        return sb;
    }
    
    static void access$100(StandardGlyphVector standardGlyphVector, int n) {
        standardGlyphVector.clearCaches(n);
    }
    
    static void access$200(StandardGlyphVector standardGlyphVector, int n) {
        standardGlyphVector.clearFlags(n);
    }
    
    static void access$300(StandardGlyphVector standardGlyphVector, int n) {
        standardGlyphVector.addFlags(n);
    }
    
    static StandardGlyphVector.GlyphStrike access$500(StandardGlyphVector standardGlyphVector) {
        return standardGlyphVector.getDefaultStrike();
    }
    
    static final class GlyphTransformInfo
    {
        StandardGlyphVector sgv;
        int[] indices;
        double[] transforms;
        SoftReference strikesRef;
        boolean haveAllStrikes;
        
        GlyphTransformInfo(StandardGlyphVector sgv) {
            this.sgv = sgv;
        }
        
        GlyphTransformInfo(StandardGlyphVector sgv, StandardGlyphVector.GlyphTransformInfo glyphTransformInfo) {
            this.sgv = sgv;
            this.indices = ((glyphTransformInfo.indices == null) ? null : ((int[])(int[])glyphTransformInfo.indices.clone()));
            this.transforms = ((glyphTransformInfo.transforms == null) ? null : ((double[])(double[])glyphTransformInfo.transforms.clone()));
            this.strikesRef = null;
        }
        
        public boolean equals(StandardGlyphVector.GlyphTransformInfo glyphTransformInfo) {
            if (glyphTransformInfo == null) {
                return false;
            }
            if (glyphTransformInfo == this) {
                return true;
            }
            if (this.indices.length != glyphTransformInfo.indices.length) {
                return false;
            }
            if (this.transforms.length != glyphTransformInfo.transforms.length) {
                return false;
            }
            int i = 0;
            while (i < this.indices.length) {
                int n = this.indices[i];
                int n2 = glyphTransformInfo.indices[i];
                if (n == 0 != (n2 == 0)) {
                    return false;
                }
                if (n != 0) {
                    int n3 = n * 6;
                    int n4 = n2 * 6;
                    int j = 6;
                    while (j > 0) {
                        if (this.indices[--n3] != glyphTransformInfo.indices[--n4]) {
                            return false;
                        }
                        --j;
                    }
                }
                ++i;
            }
            return true;
        }
        
        void setGlyphTransform(int n, AffineTransform affineTransform) {
            double[] array = new double[6];
            boolean n2 = true;
            if (affineTransform == null || affineTransform.isIdentity()) {
                array[0] = (array[3] = 1.0);
            }
            else {
                n2 = false;
                affineTransform.getMatrix(array);
            }
            if (this.indices == null) {
                if (!(n2 == 0)) {
                    return;
                }
                (this.indices = new int[this.sgv.glyphs.length])[n] = 1;
                this.transforms = array;
            }
            else {
                boolean n3 = false;
                int n4;
                if (!(n2 == 0)) {
                    n4 = 0;
                }
                else {
                    n3 = true;
                    int i = 0;
                Label_0103:
                    while (i < this.transforms.length) {
                        int j = 0;
                        while (j < 6) {
                            if (this.transforms[i + j] != array[j]) {
                                i += 6;
                                continue Label_0103;
                            }
                            ++j;
                        }
                        n3 = false;
                        break;
                    }
                    n4 = i / 6 + 1;
                }
                int n5 = this.indices[n];
                if (n4 != n5) {
                    boolean n6 = false;
                    if (n5 != 0) {
                        n6 = true;
                        int k = 0;
                        while (k < this.indices.length) {
                            if (this.indices[k] == n5 && k != n) {
                                n6 = false;
                                break;
                            }
                            ++k;
                        }
                    }
                    if (n6 && n3) {
                        n4 = n5;
                        System.arraycopy(array, 0, this.transforms, (n4 - 1) * 6, 6);
                    }
                    else if (!(n6 == 0)) {
                        if (this.transforms.length == 6) {
                            this.indices = null;
                            this.transforms = null;
                            StandardGlyphVector.access$100(this.sgv, n);
                            StandardGlyphVector.access$200(this.sgv, 1);
                            this.strikesRef = null;
                            return;
                        }
                        double[] transforms2 = new double[this.transforms.length - 6];
                        System.arraycopy(this.transforms, 0, transforms2, 0, (n5 - 1) * 6);
                        System.arraycopy(this.transforms, n5 * 6, transforms2, (n5 - 1) * 6, this.transforms.length - n5 * 6);
                        this.transforms = transforms2;
                        for (int l = 0; l < this.indices.length; ++l) {
                            if (this.indices[l] > n5) {
                                int[] indices = this.indices;
                                int n7 = l;
                                --indices[n7];
                            }
                        }
                        if (n4 > n5) {
                            --n4;
                        }
                    }
                    else if (!(n3 == 0)) {
                        double[] transforms = new double[this.transforms.length + 6];
                        System.arraycopy(this.transforms, 0, transforms, 0, this.transforms.length);
                        System.arraycopy(array, 0, transforms, this.transforms.length, 6);
                        this.transforms = transforms;
                    }
                    this.indices[n] = n4;
                }
            }
            StandardGlyphVector.access$100(this.sgv, n);
            StandardGlyphVector.access$300(this.sgv, 1);
            this.strikesRef = null;
        }
        
        AffineTransform getGlyphTransform(int n) {
            int n2 = this.indices[n];
            if (n2 == 0) {
                return null;
            }
            int n3 = (n2 - 1) * 6;
            return new AffineTransform(this.transforms[n3 + 0], this.transforms[n3 + 1], this.transforms[n3 + 2], this.transforms[n3 + 3], this.transforms[n3 + 4], this.transforms[n3 + 5]);
        }
        
        int transformCount() {
            if (this.transforms == null) {
                return 0;
            }
            return this.transforms.length / 6;
        }
        
        Object setupGlyphImages(long[] array, float[] array2, AffineTransform affineTransform) {
            int length = this.sgv.glyphs.length;
            StandardGlyphVector.GlyphStrike[] allStrikes = this.getAllStrikes();
            for (int i = 0; i < length; ++i) {
                StandardGlyphVector.GlyphStrike glyphStrike = allStrikes[this.indices[i]];
                int n = this.sgv.glyphs[i];
                array[i] = glyphStrike.strike.getGlyphImagePtr(n);
                glyphStrike.getGlyphPosition(n, i * 2, this.sgv.positions, array2);
            }
            affineTransform.transform(array2, 0, array2, 0, length);
            return allStrikes;
        }
        
        Rectangle getGlyphsPixelBounds(AffineTransform affineTransform, float n, float n2, int n3, int n4) {
            Object o = null;
            Rectangle rectangle = new Rectangle();
            Point2D.Float float1 = new Point2D.Float();
            int n5 = n3 * 2;
            while (--n4 >= 0) {
                StandardGlyphVector.GlyphStrike strike = this.getStrike(n3);
                float1.x = n + this.sgv.positions[n5++] + strike.dx;
                float1.y = n2 + this.sgv.positions[n5++] + strike.dy;
                affineTransform.transform(float1, float1);
                strike.strike.getGlyphImageBounds(this.sgv.glyphs[n3++], float1, rectangle);
                if (!rectangle.isEmpty()) {
                    if (o == null) {
                        o = new Rectangle(rectangle);
                    }
                    else {
                        o.add(rectangle);
                    }
                }
                else {
                    continue;
                }
            }
            return (o != null) ? o : rectangle;
        }
        
        StandardGlyphVector.GlyphStrike getStrike(int n) {
            if (this.indices != null) {
                return this.getStrikeAtIndex(this.getStrikeArray(), this.indices[n]);
            }
            return StandardGlyphVector.access$500(this.sgv);
        }
        
        private StandardGlyphVector.GlyphStrike[] getAllStrikes() {
            if (this.indices == null) {
                return null;
            }
            StandardGlyphVector.GlyphStrike[] strikeArray = this.getStrikeArray();
            if (!this.haveAllStrikes) {
                for (int i = 0; i < strikeArray.length; ++i) {
                    this.getStrikeAtIndex(strikeArray, i);
                }
                this.haveAllStrikes = true;
                return strikeArray;
            }
            return strikeArray;
        }
        
        private StandardGlyphVector.GlyphStrike[] getStrikeArray() {
            StandardGlyphVector.GlyphStrike[] referent = null;
            if (this.strikesRef != null) {
                referent = (StandardGlyphVector.GlyphStrike[])(StandardGlyphVector.GlyphStrike[])this.strikesRef.get();
            }
            if (referent == null) {
                this.haveAllStrikes = false;
                referent = new StandardGlyphVector.GlyphStrike[this.transformCount() + 1];
                this.strikesRef = new SoftReference(referent);
                return referent;
            }
            return referent;
        }
        
        private StandardGlyphVector.GlyphStrike getStrikeAtIndex(StandardGlyphVector.GlyphStrike[] array, int n) {
            StandardGlyphVector.GlyphStrike glyphStrike = array[n];
            if (glyphStrike == null) {
                if (n == 0) {
                    glyphStrike = StandardGlyphVector.access$500(this.sgv);
                }
                else {
                    int n2 = (n - 1) * 6;
                    glyphStrike = StandardGlyphVector.GlyphStrike.create(this.sgv, this.sgv.dtx, new AffineTransform(this.transforms[n2], this.transforms[n2 + 1], this.transforms[n2 + 2], this.transforms[n2 + 3], this.transforms[n2 + 4], this.transforms[n2 + 5]));
                }
                array[n] = glyphStrike;
                return glyphStrike;
            }
            else {
                return glyphStrike;
            }
        }
    }
    
    public static final class GlyphStrike
    {
        StandardGlyphVector sgv;
        FontStrike strike;
        float dx;
        float dy;
        
        static StandardGlyphVector.GlyphStrike create(StandardGlyphVector standardGlyphVector, AffineTransform tx, AffineTransform tx2) {
            float n = 0.0f;
            float n2 = 0.0f;
            AffineTransform access$700 = standardGlyphVector.ftx;
            if (!tx.isIdentity() || tx2 != null) {
                access$700 = new AffineTransform(standardGlyphVector.ftx);
                if (tx2 != null) {
                    access$700.preConcatenate(tx2);
                    n = (float)access$700.getTranslateX();
                    n2 = (float)access$700.getTranslateY();
                }
                if (!tx.isIdentity()) {
                    access$700.preConcatenate(tx);
                }
            }
            int n3 = 1;
            Object antiAliasingHint = standardGlyphVector.frc.getAntiAliasingHint();
            if (antiAliasingHint == RenderingHints.VALUE_TEXT_ANTIALIAS_GASP && !access$700.isIdentity() && (access$700.getType() & 0xFFFFFFFE) != 0x0) {
                double shearX = access$700.getShearX();
                if (shearX != 0.0) {
                    double scaleY = access$700.getScaleY();
                    n3 = (int)Math.sqrt(shearX * shearX + scaleY * scaleY);
                }
                else {
                    n3 = (int)Math.abs(access$700.getScaleY());
                }
            }
            FontStrikeDesc o = new FontStrikeDesc(tx, access$700, standardGlyphVector.font.getStyle(), FontStrikeDesc.getAAHintIntVal(antiAliasingHint, standardGlyphVector.font2D, n3), FontStrikeDesc.getFMHintIntVal(standardGlyphVector.frc.getFractionalMetricsHint()));
            Font2D font2D = standardGlyphVector.font2D;
            if ((font2D instanceof FontSubstitution)) {
                font2D = ((FontSubstitution)font2D).getCompositeFont2D();
            }
            return new StandardGlyphVector.GlyphStrike(standardGlyphVector, font2D.handle.font2D.getStrike(o), n, n2);
        }
        
        private GlyphStrike(StandardGlyphVector sgv, FontStrike strike, float dx, float dy) {
            this.sgv = sgv;
            this.strike = strike;
            this.dx = dx;
            this.dy = dy;
        }
        
        void getADL(StandardGlyphVector.ADL adl) {
            StrikeMetrics fontMetrics = this.strike.getFontMetrics();
            if (this.sgv.font.isTransformed()) {
                Point2D.Float float1 = new Point2D.Float();
                float1.x = (float)this.sgv.font.getTransform().getTranslateX();
                float1.y = (float)this.sgv.font.getTransform().getTranslateY();
            }
            adl.ascentX = -fontMetrics.ascentX;
            adl.ascentY = -fontMetrics.ascentY;
            adl.descentX = fontMetrics.descentX;
            adl.descentY = fontMetrics.descentY;
            adl.leadingX = fontMetrics.leadingX;
            adl.leadingY = fontMetrics.leadingY;
        }
        
        void getGlyphPosition(int n, int n2, float[] array, float[] array2) {
            array2[n2] = array[n2] + this.dx;
            ++n2;
            array2[n2] = array[n2] + this.dy;
        }
        
        void addDefaultGlyphAdvance(int n, Point2D.Float float1) {
            Point2D.Float glyphMetrics = this.strike.getGlyphMetrics(n);
            float1.x += glyphMetrics.x + this.dx;
            float1.y += glyphMetrics.y + this.dy;
        }
        
        Rectangle2D getGlyphOutlineBounds(int n, float n2, float n3) {
            Rectangle2D.Float bounds2D;
            if (this.sgv.invdtx == null) {
                bounds2D = new Rectangle2D.Float();
                bounds2D.setRect(this.strike.getGlyphOutlineBounds(n));
            }
            else {
                GeneralPath glyphOutline = this.strike.getGlyphOutline(n, 0.0f, 0.0f);
                glyphOutline.transform(this.sgv.invdtx);
                bounds2D = glyphOutline.getBounds2D();
            }
            if (!bounds2D.isEmpty()) {
                bounds2D.setRect(bounds2D.getMinX() + (double)n2 + (double)this.dx, bounds2D.getMinY() + (double)n3 + (double)this.dy, bounds2D.getWidth(), bounds2D.getHeight());
                return bounds2D;
            }
            return bounds2D;
        }
        
        void appendGlyphOutline(int n, GeneralPath generalPath, float n2, float n3) {
            GeneralPath generalPath2;
            if (this.sgv.invdtx == null) {
                generalPath2 = this.strike.getGlyphOutline(n, n2 + this.dx, n3 + this.dy);
            }
            else {
                generalPath2 = this.strike.getGlyphOutline(n, 0.0f, 0.0f);
                generalPath2.transform(this.sgv.invdtx);
                generalPath2.transform(AffineTransform.getTranslateInstance((double)(n2 + this.dx), (double)(n3 + this.dy)));
            }
            generalPath.append(generalPath2.getPathIterator(null), false);
        }
    }
    
    static class ADL
    {
        public float ascentX;
        public float ascentY;
        public float descentX;
        public float descentY;
        public float leadingX;
        public float leadingY;
        
        @Override
        public String toString() {
            return this.toStringBuffer(null).toString();
        }
        
        protected StringBuffer toStringBuffer(StringBuffer sb) {
            if (sb == null) {
                sb = new StringBuffer();
            }
            sb.append("ax: ");
            sb.append(this.ascentX);
            sb.append(" ay: ");
            sb.append(this.ascentY);
            sb.append(" dx: ");
            sb.append(this.descentX);
            sb.append(" dy: ");
            sb.append(this.descentY);
            sb.append(" lx: ");
            sb.append(this.leadingX);
            sb.append(" ly: ");
            sb.append(this.leadingY);
            return sb;
        }
    }
}
